'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var inside = require('is-path-inside');
var upath = require('upath');
var chalk = require('chalk');
var consola = require('consola');
var core = require('@midwayjs/core');
var tslib = require('tslib');
var decorator = require('@midwayjs/decorator');
var asynchronousLocalStorage = require('asynchronous-local-storage');
var path = require('path');
var pkgDir = require('pkg-dir');
var createJITI = require('jiti');
var _ = require('lodash');
var staticCache = require('koa-static-cache');
var fs = require('fs');
var superjson = require('superjson');
var request_dist_midwayjsHooksCoreRequest = require('../request/dist/midwayjs-hooks-core-request.cjs.prod.js');
var esModuleLexer = require('es-module-lexer');
require('axios');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var inside__default = /*#__PURE__*/_interopDefault(inside);
var chalk__default = /*#__PURE__*/_interopDefault(chalk);
var consola__default = /*#__PURE__*/_interopDefault(consola);
var path__default = /*#__PURE__*/_interopDefault(path);
var createJITI__default = /*#__PURE__*/_interopDefault(createJITI);
var ___default = /*#__PURE__*/_interopDefault(_);
var staticCache__default = /*#__PURE__*/_interopDefault(staticCache);
var fs__default = /*#__PURE__*/_interopDefault(fs);

function isProduction() {
  if (process.env.MIDWAY_TS_MODE === 'true' ||         "production" === 'test' ||         "production" === 'development') {
    return false;
  }

  return true;
}

class ServerRouter {
  constructor(root, config, useSource = true) {
    this.root = void 0;
    this.config = void 0;
    this.routes = new Map();
    this.useSource = void 0;
    this.root = root;
    this.config = config;
    this.useSource = useSource;
  } // src/apis


  get source() {
    if (this.useSource) {
      return upath.join(this.root, this.config.source);
    }

    return upath.join(this.root, this.config.build.outDir);
  }

  getRouteConfig(file) {
    const {
      routes
    } = this.config;
    const index = routes.findIndex(route => {
      const dir = this.getApiDirectory(route.baseDir);
      return this.inside(file, dir);
    });

    if (!routes[index]) {
      return null;
    }

    return {
      underscore: false,
      ...routes[index]
    };
  } // src/apis/lambda


  getApiDirectory(baseDir) {
    return upath.join(this.source, baseDir);
  }

  isApiFile(file) {
    if (file.endsWith('.test.ts') || file.endsWith('.test.js')) {
      return false;
    }

    if (upath.extname(file) !== '.ts' && upath.extname(file) !== '.js') {
      return false;
    }

    const route = this.getRouteConfig(file);
    return !!route;
  }

  inside(child, parent) {
    return inside__default['default'](upath.toUnix(child), upath.toUnix(parent));
  }

  getBaseUrl(file) {
    const url = this.getHTTPPath(file, '', true);

    if (url === '/*') {
      return '/';
    }

    if (url.endsWith('/*')) {
      return url.slice(0, -2);
    }

    return url;
  }

  getHTTPPath(file, method, isExportDefault) {
    const {
      basePath,
      baseDir,
      underscore
    } = this.getRouteConfig(file);
    const lambdaDirectory = this.getApiDirectory(baseDir);
    const {
      isCatchAllRoutes,
      filename
    } = parseFilename(upath.basename(file, upath.extname(file)));
    const fileRoute = filename === 'index' ? '' : filename; // TODO remove underscore support in future

    const ApiMethodPrefix = '_';
    const methodPrefix = underscore ? ApiMethodPrefix : '';
    const func = isExportDefault ? '' : `${methodPrefix}${method}`;
    const api = upath.toUnix(upath.join(basePath,
    /**
     * /apis/lambda/index.ts -> ''
     * /apis/lambda/todo/index.ts -> 'todo'
     */
    upath.relative(lambdaDirectory, upath.dirname(file)),
    /**
     * index -> ''
     * demo -> '/demo'
     */
    fileRoute, // getTodoList -> _getTodoList
    func, isCatchAllRoutes ? '/*' : ''));
    /**
     * duplicate routes
     */

    const existPath = this.routes.get(api);

    if (existPath && existPath !== upath.toUnix(file)) {
      duplicateLogger(this.root, existPath, file, api);
    }

    this.routes.set(api, upath.toUnix(file));
    return api;
  }

}

function parseFilename(filename) {
  var _re$exec;

  const re = /\[\.{3}(.+)]/;
  const isCatchAllRoutes = re.test(filename);
  return {
    isCatchAllRoutes,
    filename: isCatchAllRoutes ? (_re$exec = re.exec(filename)) === null || _re$exec === void 0 ? void 0 : _re$exec[1] : filename
  };
}

function duplicateLogger(root, existPath, currentPath, api) {
  consola__default['default'].info('[ %s ] Duplicate routes detected. %s and %s both resolve to %s. Reference: %s', chalk__default['default'].yellow('warn'), chalk__default['default'].cyan(upath.relative(root, existPath)), chalk__default['default'].cyan(upath.relative(root, currentPath)), chalk__default['default'].cyan(api), 'https://www.yuque.com/midwayjs/faas/et7x4k');
}

/**
 * @private
 * private api, may change without notice.
 */
const als = {
  getStore(key) {
    return asynchronousLocalStorage.als.get(key);
  },

  run(ctx, callback) {
    return new Promise(resolve => {
      asynchronousLocalStorage.als.runWith(async () => {
        resolve(await callback());
      }, ctx);
    });
  }

}; // /**
//  * @private
//  * private api, may change without notice.
//  */
// export const als = new AsyncLocalStorage<HooksContext>()

function getProjectRoot(cwd) {
  return pkgDir.sync(cwd) || process.cwd();
}
function getConfig(cwd) {
  const root = getProjectRoot(cwd);
  const configs = {
    ts: path__default['default'].join(root, 'midway.config.ts'),
    js: path__default['default'].join(root, 'midway.config.js')
  };
  const userConfig = tryRequire(configs.ts) || tryRequire(configs.js);
  return ___default['default'].defaultsDeep({}, userConfig, {
    source: './src/apis',
    build: {
      viteOutDir: './build',
      outDir: './dist'
    }
  });
}
function defineConfig(config) {
  return config;
}

const tryRequire = id => {
  try {
    const jiti = createJITI__default['default']();
    const contents = jiti(id);
    if ('default' in contents) return contents.default;
    return contents;
  } catch {
    return undefined;
  }
};

/**
 * Create hooks component
 */

const hooks = (config = {}) => {
  return new HooksComponent(config).createConfiguration();
};

class HooksComponent {
  constructor(componentConfig) {
    this.root = void 0;
    this.config = void 0;
    this.componentConfig = void 0;
    this.router = void 0;
    this.container = void 0;
    this.componentConfig = componentConfig;
    this.root = getProjectRoot();
    this.config = getConfig();
    this.router = new ServerRouter(this.root, this.config, !isProduction());
  }

  createConfiguration() {
    let count = 0;
    const configuration = core.createConfiguration({
      namespace: '@midwayjs/hooks',
      directoryResolveFilter: this.config.routes.map((route, index) => {
        return {
          pattern: route.baseDir,
          ignoreRequire: true,
          filter: (_, file, container) => {
            if (!this.container) this.container = container;
            if (!this.router.isApiFile(file)) return;
            this.createApi(file);
            count++;

            if (count === this.config.routes.length) {
              this.createRender();
            }
          }
        };
      })
    });
    configuration.onReady((_, app) => {
      this.applyMiddleware(app);
    }).onStop(_.noop);
    return {
      Configuration: configuration
    };
  }

  createApi(file) {
    var _mod$config;

    const mod = require(file);

    const modMiddleware = (mod === null || mod === void 0 ? void 0 : (_mod$config = mod.config) === null || _mod$config === void 0 ? void 0 : _mod$config.middleware) || [];
    Object.keys(mod).filter(key => typeof mod[key] === 'function').forEach(key => {
      this.createFunction({
        fn: mod[key],
        file: file,
        isExportDefault: key === 'default',
        modMiddleware
      });
    });
  }

  createFunction(config) {
    const {
      fn,
      file,
      isExportDefault,
      modMiddleware
    } = config;
    const fnName = isExportDefault ? '$default' : fn.name;
    const id = this.getFunctionId(file, fnName, isExportDefault);
    const containerId = 'hooks::' + id;
    const httpPath = this.router.getHTTPPath(file, fnName, isExportDefault);
    const httpMethod = fn.length === 0 ? 'GET' : 'POST'; // Set param for unit testing

    fn._param = {
      url: httpPath,
      method: httpMethod,
      meta: {
        functionName: id
      }
    } // Apply module middleware
    ;
    (fn.middleware || (fn.middleware = [])).unshift(...modMiddleware);
    fn.middleware = fn.middleware.map(middleware => covert(middleware));
    this.registerFunctionToContainer({
      containerId,
      httpMethod,
      httpPath,
      fn
    });
  }

  registerFunctionToContainer(config) {
    var _temp;

    const {
      containerId,
      httpMethod,
      httpPath,
      fn
    } = config;
    const Method = httpMethod === 'GET' ? decorator.Get : decorator.Post; // Source: https://shorturl.at/pqI06

    let FunctionContainer = (_temp = class FunctionContainer {
      constructor() {
        this.ctx = void 0;
      }

      async handler() {
        var _this$ctx$request, _this$ctx$request$bod;

        let args = ((_this$ctx$request = this.ctx.request) === null || _this$ctx$request === void 0 ? void 0 : (_this$ctx$request$bod = _this$ctx$request.body) === null || _this$ctx$request$bod === void 0 ? void 0 : _this$ctx$request$bod.args) || [];

        if (typeof args === 'string') {
          args = JSON.parse(args);
        }

        return superjson.serialize(await fn(...args));
      }

    }, _temp);

    tslib.__decorate([decorator.Inject()], FunctionContainer.prototype, 'ctx', void 0);

    tslib.__decorate([Method(httpPath, {
      middleware: fn.middleware
    })], FunctionContainer.prototype, 'handler', null);

    FunctionContainer = tslib.__decorate([decorator.Provide(containerId), decorator.Controller('/')], FunctionContainer);
    this.container.bind(containerId, FunctionContainer);
  }

  createRender() {
    // Only available  in production and does not register /* routes
    if (!isProduction() || !this.isFullStackProject()) {
      return;
    }

    if (this.existRootPath) {
      consola__default['default'].info('The route `/` or `/*` is registered, you need to host the front-end page manually');
      return;
    }

    this.registerFunctionToContainer({
      containerId: 'hooks:render',
      httpMethod: 'GET',
      httpPath: '/*',
      fn: async () => {}
    });
  }

  get existRootPath() {
    return this.router.routes.has('/') || this.router.routes.has('/*');
  }

  applyMiddleware(app) {
    app.use(async (ctx, next) => {
      await als.run({
        ctx
      }, async () => await next());
    }); // Apply global middleware from config

    if (Array.isArray(this.componentConfig.middleware)) {
      this.componentConfig.middleware.forEach(middleware => app.use(covert(middleware)));
    } // Serve vite static html


    const type = app.getFrameworkType();
    const requireStaticCache = type === core.MidwayFrameworkType.WEB_KOA || type === core.MidwayFrameworkType.FAAS;

    if (isProduction() && requireStaticCache && this.isFullStackProject() && !this.existRootPath) {
      const baseDir = app.getBaseDir();
      app.use(staticCache__default['default']({
        dir: path.join(baseDir, '..', this.config.build.viteOutDir),
        dynamic: true,
        alias: {
          '/': 'index.html'
        },
        buffer: true,
        gzip: true
      }));
    }
  }

  getFunctionId(file, functionName, isExportDefault) {
    const rule = this.router.getRouteConfig(file);
    const lambdaDirectory = this.router.getApiDirectory(rule.baseDir);
    const length = this.router.config.routes.length; // 多个 source 的情况下，根据各自的 lambdaDirectory 来增加前缀命名

    const relativeDirectory = length > 1 ? this.router.source : lambdaDirectory;
    const relativePath = upath.relative(relativeDirectory, file); // a/b/c -> a-b-c

    const id = _.kebabCase(upath.removeExt(relativePath, upath.extname(relativePath)));
    const name = [id, isExportDefault ? '' : `-${functionName}`].join('');
    return name.toLowerCase();
  } // TODO Refactor to use config


  isFullStackProject() {
    const configs = ['vite.config.ts', 'vite.config.js'];
    return configs.some(config => fs__default['default'].existsSync(path.join(this.root, config)));
  }

}

function covert(fn) {
  return (...args) => {
    /**
     * Hooks middleware
     * const middleware = (next) => { const ctx = useContext() }
     */
    if (fn.length === 1) {
      return fn(args.pop());
    }

    return fn(...args);
  };
}

async function parseAndGenerateSDK(baseUrl, code) {
  await esModuleLexer.init;
  const [, exports] = esModuleLexer.parse(code);

  if (exports.length === 0) {
    return null;
  }

  return `
import { createRequest } from '@midwayjs/hooks-core/request';

${exports.map(id => {
    if (id === 'default') {
      return `export default createRequest('${baseUrl}', '${id}');`;
    }

    return `export const ${id} = createRequest('${baseUrl}', '${id}');`;
  }).join('\n')}
  `.trim();
}

Object.defineProperty(exports, 'consola', {
  enumerable: true,
  get: function () {
    return consola__default['default'];
  }
});
exports.createRequest = request_dist_midwayjsHooksCoreRequest.createRequest;
exports.defaults = request_dist_midwayjsHooksCoreRequest.defaults;
exports.request = request_dist_midwayjsHooksCoreRequest.request;
exports.ServerRouter = ServerRouter;
exports.als = als;
exports.defineConfig = defineConfig;
exports.getConfig = getConfig;
exports.getProjectRoot = getProjectRoot;
exports.hooks = hooks;
exports.isProduction = isProduction;
exports.parseAndGenerateSDK = parseAndGenerateSDK;
